import pytest
from terminusdb_client.woqlclient.connectionCapabilities import ConnectionCapabilities

from .AllServerRecords import AllServerRecords
from .connectCapabilitiesResponse import ConnectResponse
from .connectionObjDump import ConnectionDump
from .DBRecord import db_record_obj
from .serverRecordsFromCap import server_records_from_cap

url = "http://localhost:6363/"
json_context = {
    "layer": "http://terminusdb.com/schema/layer#",
    "owl": "http://www.w3.org/2002/07/owl#",
    "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
    "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
    "ref": "http://terminusdb.com/schema/ref#",
    "repo": "http://terminusdb.com/schema/repository#",
    "terminus": "http://terminusdb.com/schema/terminus#",
    "vio": "http://terminusdb.com/schema/vio#",
    "woql": "http://terminusdb.com/schema/woql#",
    "xdd": "http://terminusdb.com/schema/xdd#",
    "xsd": "http://www.w3.org/2001/XMLSchema#",
}

dbrec = {
    "db": "aaaaaa",
    "account": "admin",
    "title": "admin|aaaaaa",
    "description": "dasdasdds",
}

dbrecs = [
    {"db": "5534534", "account": "admin", "title": "3453", "description": "345345"},
    {
        "db": "adsasasddsa",
        "account": "admin",
        "title": "admin|adsasasddsa",
        "description": "asdsadsda",
    },
    {"db": "blah", "account": "admin", "title": "admin|blah", "description": "adfadf"},
    {
        "db": "daassd",
        "account": "admin",
        "title": "admin|daassd",
        "description": "asdasdsd",
    },
    {
        "db": "dassadds",
        "account": "admin",
        "title": "admin|dassadds",
        "description": "asdsdsad",
    },
    {"db": "ddd", "account": "admin", "title": "admin|ddd", "description": "ddds"},
    {"db": "ffff", "account": "admin", "title": "admin|ffff", "description": "fffff"},
    {
        "db": "fffffff",
        "account": "admin",
        "title": "admin|fffffff",
        "description": "fff",
    },
    {
        "db": "ggggggggg",
        "account": "admin",
        "title": "admin|ggggggggg",
        "description": "gg",
    },
    {
        "db": "twretwert",
        "account": "admin",
        "title": "admin|twretwert",
        "description": "adfadf",
    },
    {
        "db": "terminus",
        "account": "",
        "title": "Master Database",
        "description": "The master database contains the meta-data about databases, users and roles",
    },
]

connection_capabilities = ConnectionCapabilities()
connection_capabilities.set_capabilities(ConnectResponse)


class TestCapabilitiesActions:
    def test_connection_capability_object(self):
        assert connection_capabilities.connection == ConnectionDump

    def test_form_resource_name(self):
        assert (
            connection_capabilities._form_resource_name("aaaaaa", "admin")
            == "admin|aaaaaa"
        )
        assert (
            connection_capabilities.find_resource_document_id("aaaaaa", "admin")
            == "doc:Database%5fadmin%7Caaaaaa"
        )

    def test_get_server_record(self):
        x = connection_capabilities._get_server_record()
        assert x == server_records_from_cap

    def test_get_json_context(self):
        assert connection_capabilities.get_json_context() == json_context

    def test_capabilities_permit(self):
        assert connection_capabilities.capabilities_permit(
            "create_database", "aaaaaa", "admin"
        )
        with pytest.raises(Exception):
            assert connection_capabilities.capabilities_permit(
                "delete_database", "nonexistant", "admin"
            )

    def test_get_db_record(self):
        assert (
            connection_capabilities._get_db_record("aaaaaa", "admin") == db_record_obj
        )

    def test_extract_metadata(self):
        assert connection_capabilities._extract_metadata(db_record_obj) == dbrec

    def test_get_db_metadata(self):
        assert connection_capabilities._get_db_metadata("aaaaaa", "admin") == dbrec

    def test_remove_db(self):
        connection_capabilities.remove_db("aaaaaa", "admin")
        with pytest.raises(Exception):
            assert connection_capabilities._get_db_metadata("aaaaaa", "admin")

    def test_get_server_db_records(self):
        x = connection_capabilities.get_server_db_records()
        assert x == AllServerRecords

    def test_get_server_db_metadata(self):
        x = connection_capabilities.get_server_db_metadata()
        assert x == dbrecs



# import pprint

from terminusdb_client.woqlquery.woql_query import WOQLQuery

from .woqljson.woqlExtraJson import WoqlExtra

# pp = pprint.PrettyPrinter(indent=4)


class TestWoqlExtra:
    def test_using(self):
        woql_object = WOQLQuery().using(
            "userName/dbName/local/commit/commitID",
            WOQLQuery().triple("v:A", "v:B", "v:C"),
        )
        woql_triple = WOQLQuery().triple("v:A", "v:B", "v:C")
        woql_object01 = WOQLQuery().using(
            "userName/dbName/local/commit/commitID", woql_triple
        )
        assert woql_object.to_dict() == WoqlExtra["usingJson"]
        assert woql_object01.to_dict() == WoqlExtra["usingJson"]

    def test_multi_using(self):
        woql_object = WOQLQuery()
        woql_object.woql_and(
            WOQLQuery().using(
                "admin/dbName/local/commit/commitID_1",
                WOQLQuery().triple("v:A", "v:B", "v:C"),
            ),
            WOQLQuery().using(
                "admin/dbName/local/commit/commitID_2",
                WOQLQuery().woql_not(WOQLQuery().triple("v:A", "v:B", "v:C")),
            ),
        )
        assert woql_object.to_dict() == WoqlExtra["multiUsingJson"]

    def test_chain_and(self):
        woql_query = WOQLQuery()
        woql_query.woql_and(
            WOQLQuery().triple("v:A", "v:B", "v:C"),
            WOQLQuery().triple("v:D", "v:E", "v:F"),
        )
        assert woql_query.to_dict() == WoqlExtra["chainAndJson"]



class APIEndpointConst:
    CONNECT = "connect"
    GET_TRIPLES = "get_triples"
    UPDATE_TRIPLES = "update_triples"
    CLASS_FRAME = "class_frame"
    WOQL_QUERY = "woql_query"
    DELETE_DATABASE = "delete_database"
    CREATE_DATABASE = "create_database"
    CREATE_GRAPH = ("create_graph",)
    DELETE_GRAPH = ("delete_graph",)
    CLONE = ("clone",)
    PUSH = ("push",)
    FETCH = ("fetch",)
    PULL = ("pull",)
    REBASE = ("rebase",)
    BRANCH = "branch"



AllServerRecords = {
    "doc:Database%5fadmin%7C5534534": {
        "@id": "doc:Database%5fadmin%7C5534534",
        "@type": "terminus:Database",
        "rdfs:comment": {"@language": "en", "@value": "345345"},
        "rdfs:label": [
            {"@language": "en", "@value": "3453"},
            {"@language": "en", "@value": "admin|5534534"},
        ],
        "terminus:allow_origin": {"@type": "xsd:string", "@value": "*"},
        "terminus:database_state": {
            "@id": "terminus:finalized",
            "@type": "terminus:DatabaseState",
        },
        "terminus:resource_name": {"@type": "xsd:string", "@value": "admin|5534534"},
        "terminus:authority": [
            "terminus:branch",
            "terminus:class_frame",
            "terminus:clone",
            "terminus:commit_read_access",
            "terminus:commit_write_access",
            "terminus:create_database",
            "terminus:delete_database",
            "terminus:fetch",
            "terminus:inference_read_access",
            "terminus:inference_write_access",
            "terminus:instance_read_access",
            "terminus:instance_write_access",
            "terminus:manage",
            "terminus:meta_read_access",
            "terminus:meta_write_access",
            "terminus:push",
            "terminus:rebase",
            "terminus:schema_read_access",
            "terminus:schema_write_access",
        ],
    },
    "doc:Database%5fadmin%7Cadsasasddsa": {
        "@id": "doc:Database%5fadmin%7Cadsasasddsa",
        "@type": "terminus:Database",
        "rdfs:comment": {"@language": "en", "@value": "asdsadsda"},
        "rdfs:label": [
            {"@language": "en", "@value": "admin|adsasasddsa"},
            {"@language": "en", "@value": "asdasdasd"},
        ],
        "terminus:allow_origin": {"@type": "xsd:string", "@value": "*"},
        "terminus:database_state": {
            "@id": "terminus:finalized",
            "@type": "terminus:DatabaseState",
        },
        "terminus:resource_name": {
            "@type": "xsd:string",
            "@value": "admin|adsasasddsa",
        },
        "terminus:authority": [
            "terminus:branch",
            "terminus:class_frame",
            "terminus:clone",
            "terminus:commit_read_access",
            "terminus:commit_write_access",
            "terminus:create_database",
            "terminus:delete_database",
            "terminus:fetch",
            "terminus:inference_read_access",
            "terminus:inference_write_access",
            "terminus:instance_read_access",
            "terminus:instance_write_access",
            "terminus:manage",
            "terminus:meta_read_access",
            "terminus:meta_write_access",
            "terminus:push",
            "terminus:rebase",
            "terminus:schema_read_access",
            "terminus:schema_write_access",
        ],
    },
    "doc:Database%5fadmin%7Cblah": {
        "@id": "doc:Database%5fadmin%7Cblah",
        "@type": "terminus:Database",
        "rdfs:comment": {"@language": "en", "@value": "adfadf"},
        "rdfs:label": [
            {"@language": "en", "@value": "admin|blah"},
            {"@language": "en", "@value": "better freaking work"},
        ],
        "terminus:allow_origin": {"@type": "xsd:string", "@value": "*"},
        "terminus:database_state": {
            "@id": "terminus:finalized",
            "@type": "terminus:DatabaseState",
        },
        "terminus:resource_name": {"@type": "xsd:string", "@value": "admin|blah"},
        "terminus:authority": [
            "terminus:branch",
            "terminus:class_frame",
            "terminus:clone",
            "terminus:commit_read_access",
            "terminus:commit_write_access",
            "terminus:create_database",
            "terminus:delete_database",
            "terminus:fetch",
            "terminus:inference_read_access",
            "terminus:inference_write_access",
            "terminus:instance_read_access",
            "terminus:instance_write_access",
            "terminus:manage",
            "terminus:meta_read_access",
            "terminus:meta_write_access",
            "terminus:push",
            "terminus:rebase",
            "terminus:schema_read_access",
            "terminus:schema_write_access",
        ],
    },
    "doc:Database%5fadmin%7Cdaassd": {
        "@id": "doc:Database%5fadmin%7Cdaassd",
        "@type": "terminus:Database",
        "rdfs:comment": {"@language": "en", "@value": "asdasdsd"},
        "rdfs:label": [
            {"@language": "en", "@value": "admin|daassd"},
            {"@language": "en", "@value": "asdsd"},
        ],
        "terminus:allow_origin": {"@type": "xsd:string", "@value": "*"},
        "terminus:database_state": {
            "@id": "terminus:finalized",
            "@type": "terminus:DatabaseState",
        },
        "terminus:resource_name": {"@type": "xsd:string", "@value": "admin|daassd"},
        "terminus:authority": [
            "terminus:branch",
            "terminus:class_frame",
            "terminus:clone",
            "terminus:commit_read_access",
            "terminus:commit_write_access",
            "terminus:create_database",
            "terminus:delete_database",
            "terminus:fetch",
            "terminus:inference_read_access",
            "terminus:inference_write_access",
            "terminus:instance_read_access",
            "terminus:instance_write_access",
            "terminus:manage",
            "terminus:meta_read_access",
            "terminus:meta_write_access",
            "terminus:push",
            "terminus:rebase",
            "terminus:schema_read_access",
            "terminus:schema_write_access",
        ],
    },
    "doc:Database%5fadmin%7Cdassadds": {
        "@id": "doc:Database%5fadmin%7Cdassadds",
        "@type": "terminus:Database",
        "rdfs:comment": {"@language": "en", "@value": "asdsdsad"},
        "rdfs:label": [
            {"@language": "en", "@value": "admin|dassadds"},
            {"@language": "en", "@value": "asdasdds"},
        ],
        "terminus:allow_origin": {"@type": "xsd:string", "@value": "*"},
        "terminus:database_state": {
            "@id": "terminus:finalized",
            "@type": "terminus:DatabaseState",
        },
        "terminus:resource_name": {"@type": "xsd:string", "@value": "admin|dassadds"},
        "terminus:authority": [
            "terminus:branch",
            "terminus:class_frame",
            "terminus:clone",
            "terminus:commit_read_access",
            "terminus:commit_write_access",
            "terminus:create_database",
            "terminus:delete_database",
            "terminus:fetch",
            "terminus:inference_read_access",
            "terminus:inference_write_access",
            "terminus:instance_read_access",
            "terminus:instance_write_access",
            "terminus:manage",
            "terminus:meta_read_access",
            "terminus:meta_write_access",
            "terminus:push",
            "terminus:rebase",
            "terminus:schema_read_access",
            "terminus:schema_write_access",
        ],
    },
    "doc:Database%5fadmin%7Cddd": {
        "@id": "doc:Database%5fadmin%7Cddd",
        "@type": "terminus:Database",
        "rdfs:comment": {"@language": "en", "@value": "ddds"},
        "rdfs:label": [
            {"@language": "en", "@value": "admin|ddd"},
            {"@language": "en", "@value": "ddd"},
        ],
        "terminus:allow_origin": {"@type": "xsd:string", "@value": "*"},
        "terminus:database_state": {
            "@id": "terminus:finalized",
            "@type": "terminus:DatabaseState",
        },
        "terminus:resource_name": {"@type": "xsd:string", "@value": "admin|ddd"},
        "terminus:authority": [
            "terminus:branch",
            "terminus:class_frame",
            "terminus:clone",
            "terminus:commit_read_access",
            "terminus:commit_write_access",
            "terminus:create_database",
            "terminus:delete_database",
            "terminus:fetch",
            "terminus:inference_read_access",
            "terminus:inference_write_access",
            "terminus:instance_read_access",
            "terminus:instance_write_access",
            "terminus:manage",
            "terminus:meta_read_access",
            "terminus:meta_write_access",
            "terminus:push",
            "terminus:rebase",
            "terminus:schema_read_access",
            "terminus:schema_write_access",
        ],
    },
    "doc:Database%5fadmin%7Cffff": {
        "@id": "doc:Database%5fadmin%7Cffff",
        "@type": "terminus:Database",
        "rdfs:comment": {"@language": "en", "@value": "fffff"},
        "rdfs:label": [
            {"@language": "en", "@value": "admin|ffff"},
            {"@language": "en", "@value": "fff"},
        ],
        "terminus:allow_origin": {"@type": "xsd:string", "@value": "*"},
        "terminus:database_state": {
            "@id": "terminus:finalized",
            "@type": "terminus:DatabaseState",
        },
        "terminus:resource_name": {"@type": "xsd:string", "@value": "admin|ffff"},
        "terminus:authority": [
            "terminus:branch",
            "terminus:class_frame",
            "terminus:clone",
            "terminus:commit_read_access",
            "terminus:commit_write_access",
            "terminus:create_database",
            "terminus:delete_database",
            "terminus:fetch",
            "terminus:inference_read_access",
            "terminus:inference_write_access",
            "terminus:instance_read_access",
            "terminus:instance_write_access",
            "terminus:manage",
            "terminus:meta_read_access",
            "terminus:meta_write_access",
            "terminus:push",
            "terminus:rebase",
            "terminus:schema_read_access",
            "terminus:schema_write_access",
        ],
    },
    "doc:Database%5fadmin%7Cfffffff": {
        "@id": "doc:Database%5fadmin%7Cfffffff",
        "@type": "terminus:Database",
        "rdfs:comment": {"@language": "en", "@value": "fff"},
        "rdfs:label": [
            {"@language": "en", "@value": "admin|fffffff"},
            {"@language": "en", "@value": "fffasdfasdf"},
        ],
        "terminus:allow_origin": {"@type": "xsd:string", "@value": "*"},
        "terminus:database_state": {
            "@id": "terminus:finalized",
            "@type": "terminus:DatabaseState",
        },
        "terminus:resource_name": {"@type": "xsd:string", "@value": "admin|fffffff"},
        "terminus:authority": [
            "terminus:branch",
            "terminus:class_frame",
            "terminus:clone",
            "terminus:commit_read_access",
            "terminus:commit_write_access",
            "terminus:create_database",
            "terminus:delete_database",
            "terminus:fetch",
            "terminus:inference_read_access",
            "terminus:inference_write_access",
            "terminus:instance_read_access",
            "terminus:instance_write_access",
            "terminus:manage",
            "terminus:meta_read_access",
            "terminus:meta_write_access",
            "terminus:push",
            "terminus:rebase",
            "terminus:schema_read_access",
            "terminus:schema_write_access",
        ],
    },
    "doc:Database%5fadmin%7Cggggggggg": {
        "@id": "doc:Database%5fadmin%7Cggggggggg",
        "@type": "terminus:Database",
        "rdfs:comment": {"@language": "en", "@value": "gg"},
        "rdfs:label": [
            {"@language": "en", "@value": "admin|ggggggggg"},
            {"@language": "en", "@value": "gg"},
        ],
        "terminus:allow_origin": {"@type": "xsd:string", "@value": "*"},
        "terminus:database_state": {
            "@id": "terminus:finalized",
            "@type": "terminus:DatabaseState",
        },
        "terminus:resource_name": {"@type": "xsd:string", "@value": "admin|ggggggggg"},
        "terminus:authority": [
            "terminus:branch",
            "terminus:class_frame",
            "terminus:clone",
            "terminus:commit_read_access",
            "terminus:commit_write_access",
            "terminus:create_database",
            "terminus:delete_database",
            "terminus:fetch",
            "terminus:inference_read_access",
            "terminus:inference_write_access",
            "terminus:instance_read_access",
            "terminus:instance_write_access",
            "terminus:manage",
            "terminus:meta_read_access",
            "terminus:meta_write_access",
            "terminus:push",
            "terminus:rebase",
            "terminus:schema_read_access",
            "terminus:schema_write_access",
        ],
    },
    "doc:Database%5fadmin%7Ctwretwert": {
        "@id": "doc:Database%5fadmin%7Ctwretwert",
        "@type": "terminus:Database",
        "rdfs:comment": {"@language": "en", "@value": "adfadf"},
        "rdfs:label": [
            {"@language": "en", "@value": "admin|twretwert"},
            {"@language": "en", "@value": "dgsdfgsg"},
        ],
        "terminus:allow_origin": {"@type": "xsd:string", "@value": "*"},
        "terminus:database_state": {
            "@id": "terminus:finalized",
            "@type": "terminus:DatabaseState",
        },
        "terminus:resource_name": {"@type": "xsd:string", "@value": "admin|twretwert"},
        "terminus:authority": [
            "terminus:branch",
            "terminus:class_frame",
            "terminus:clone",
            "terminus:commit_read_access",
            "terminus:commit_write_access",
            "terminus:create_database",
            "terminus:delete_database",
            "terminus:fetch",
            "terminus:inference_read_access",
            "terminus:inference_write_access",
            "terminus:instance_read_access",
            "terminus:instance_write_access",
            "terminus:manage",
            "terminus:meta_read_access",
            "terminus:meta_write_access",
            "terminus:push",
            "terminus:rebase",
            "terminus:schema_read_access",
            "terminus:schema_write_access",
        ],
    },
    "doc:terminus": {
        "@id": "doc:terminus",
        "@type": "terminus:Database",
        "rdfs:comment": {
            "@language": "en",
            "@value": "The master database contains the meta-data about databases, users and roles",
        },
        "rdfs:label": {"@language": "en", "@value": "Master Database"},
        "terminus:allow_origin": {"@type": "xsd:string", "@value": "*"},
        "terminus:resource_name": {"@type": "xsd:string", "@value": "terminus"},
        "terminus:authority": [
            "terminus:branch",
            "terminus:class_frame",
            "terminus:clone",
            "terminus:commit_read_access",
            "terminus:commit_write_access",
            "terminus:create_database",
            "terminus:delete_database",
            "terminus:fetch",
            "terminus:inference_read_access",
            "terminus:inference_write_access",
            "terminus:instance_read_access",
            "terminus:instance_write_access",
            "terminus:manage",
            "terminus:meta_read_access",
            "terminus:meta_write_access",
            "terminus:push",
            "terminus:rebase",
            "terminus:schema_read_access",
            "terminus:schema_write_access",
        ],
    },
}



from terminusdb_client.woqlclient.connectionConfig import ConnectionConfig


class TestConnectionConfig:
    start_server_url = "http://localhost:6363/"
    start_dbid = "testDB"
    local_user = "admin"
    # to be review !!!!!
    connection_config = ConnectionConfig(
        start_server_url,
        db=start_dbid,
        user=local_user,
        key="mykey",
        account=local_user,
    )

    db_url = "http://localhost:6363/db/admin/testDB"

    def test_get_sever_url(self):
        assert self.connection_config.server == self.start_server_url
        assert self.connection_config.db_url() == self.db_url

    def test_change_branch(self):
        self.connection_config.branch = "myBranch"
        assert self.connection_config.db_url() == self.db_url
        assert (
            self.connection_config.query_url()
            == "http://localhost:6363/woql/admin/testDB/local/branch/myBranch"
        )

    def test_change_server(self):
        new_ref = "gfhfjkflfgorpyuiioo"
        self.connection_config.ref = new_ref
        assert (
            self.connection_config.query_url()
            == "http://localhost:6363/woql/admin/testDB/local/commit/" + new_ref
        )

    def test_check_basic_auth(self):
        assert self.connection_config.basic_auth == "admin:mykey"



# from .errorMessage import ErrorMessage
from base64 import b64encode

import requests
import terminusdb_client.woql_utils as utils

from .api_endpoint_const import APIEndpointConst
from .errors import APIError


class DispatchRequest:
    def __init__(self):
        pass

    @staticmethod
    def __get_call(url, headers, payload):
        url = utils.add_params_to_url(url, payload)

        return requests.get(url, headers=headers)

    @staticmethod
    def __post_call(url, headers, payload, file_dict=None):
        if file_dict:
            return requests.post(url, json=payload, headers=headers, files=file_dict)
        else:
            headers["content-type"] = "application/json"
            return requests.post(url, json=payload, headers=headers)

    @staticmethod
    def __delete_call(url, headers, payload):
        url = utils.add_params_to_url(url, payload)
        return requests.delete(url, headers=headers)

    @staticmethod
    def __autorization_header(basic_auth=None, remote_auth=None):
        headers = {}

        # if (payload and ('terminus:user_key' in  payload)):
        # utils.encodeURIComponent(payload['terminus:user_key'])}
        if basic_auth:
            headers["Authorization"] = "Basic %s" % b64encode(
                (basic_auth).encode("utf-8")
            ).decode("utf-8")
        if remote_auth and remote_auth["type"] == "jwt":
            headers["Authorization-Remote"] = "Bearer %s" % remote_auth["key"]
        elif remote_auth and remote_auth["type"] == "basic":
            rauthstr = remote_auth["user"] + ":" + remote_auth["key"]
            headers["Authorization-Remote"] = "Basic %s" % b64encode(
                (rauthstr).encode("utf-8")
            ).decode("utf-8")
        return headers

    # url, action, payload, basic_auth, jwt=null

    @classmethod
    def send_request_by_action(
        cls,
        url,
        action,
        payload=None,
        basic_auth=None,
        remote_auth=None,
        file_dict=None,
    ):
        # payload default as empty dict is against PEP
        # print("Sending to URL____________", url)
        # print("Send Request By Action_____________", action)
        if payload is None:
            payload = {}

        try:
            request_response = None
            headers = cls.__autorization_header(basic_auth, remote_auth)

            if action in [
                APIEndpointConst.GET_TRIPLES,
                APIEndpointConst.CONNECT,
                APIEndpointConst.CLASS_FRAME,
            ]:
                request_response = cls.__get_call(url, headers, payload)

            elif action in [
                APIEndpointConst.DELETE_DATABASE,
                APIEndpointConst.DELETE_GRAPH,
            ]:
                request_response = cls.__delete_call(url, headers, payload)

            elif action in [
                APIEndpointConst.WOQL_QUERY,
                APIEndpointConst.CREATE_DATABASE,
                APIEndpointConst.UPDATE_TRIPLES,
                APIEndpointConst.CREATE_GRAPH,
                APIEndpointConst.FETCH,
                APIEndpointConst.PULL,
                APIEndpointConst.PUSH,
                APIEndpointConst.REBASE,
                APIEndpointConst.BRANCH,
                APIEndpointConst.CLONE,
            ]:
                request_response = cls.__post_call(url, headers, payload, file_dict)

            if request_response.status_code == 200:
                return request_response.json()  # if not a json not it raises an error
            else:
                # Raise an exception if a request is unsuccessful
                message = "Api Error"

                if type(request_response.text) is str:
                    message = request_response.text

                raise (
                    APIError(
                        message,
                        url,
                        request_response.json(),
                        request_response.status_code,
                    )
                )

        # to be reviewed
        # the server in the response return always content-type application/json
        except ValueError:
            # if the response type is not a json
            # print("Value Error", err)
            return request_response.text

        """
        except Exception as err:
            print(type(err))
            print(err.args)

        except requests.exceptions.RequestException as err:
            print ("Request Error",err)
        except requests.exceptions.HTTPError as err:
            print ("Http Error:",err)
        except requests.exceptions.ConnectionError as err:
            print ("Error Connecting:",err)
        except requests.exceptions.Timeout as err:
            print ("Timeout Error:",err)
        """



WOQL_TRIM_JSON = {
    "@type": "woql:Trim",
    "woql:untrimmed": {
        "@type": "woql:Datatype",
        "woql:datatype": {"@type": "xsd:string", "@value": "a"},
    },
    "woql:trimmed": {
        "@type": "woql:Datatype",
        "woql:datatype": {"@type": "xsd:string", "@value": "b"},
    },
}



# import woqlClient as woql
import numpy as np
import pytest

# import unittest.mock as mock
# from woqlclient import WOQLQuery
# from woqlclient import WOQLClient
import terminusdb_client.woqldataframe.woqlDataframe as woqlDF
from terminusdb_client.woqldataframe.woqlDataframe import EmptyException


@pytest.fixture
def mock_result():
    return {
        "bindings": [
            {
                "Product": {
                    "@type": "http://www.w3.org/2001/XMLSchema#string",
                    "@value": "STRAWBERRY CANDY",
                },
                "Qty": {
                    "@type": "http://www.w3.org/2001/XMLSchema#decimal",
                    "@value": 10,
                },
                "Stock": "http://195.201.12.87:6365/rational_warehouse/document/Stock110_101752_2019-12-22T00%3A00%3A00",
                "Warehouse": {"@language": "en", "@value": "Pittsburg"},
                "buyers_buy_product": "http://195.201.12.87:6365/rational_warehouse/document/Product101752",
                "stockDate": {
                    "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
                    "@value": "2019-12-22T00:00:00",
                },
                "warehouseCode": "http://195.201.12.87:6365/rational_warehouse/document/Warehouse110",
            },
        ],
        "graphs": [],
    }


TEST_TYPE = [
    ("http://www.w3.org/2001/XMLSchema#string", np.unicode_),
    ("http://www.w3.org/2001/XMLSchema#integer", np.int),
    ("http://www.w3.org/2001/XMLSchema#dateTime", np.datetime64),
    ("http://www.w3.org/2001/XMLSchema#decimal", np.double),
]

TEST_TYPE_VALUE = [
    ("http://www.w3.org/2001/XMLSchema#string", "10", "10"),
    ("http://www.w3.org/2001/XMLSchema#integer", "10", 10),
    (
        "http://www.w3.org/2001/XMLSchema#dateTime",
        "2019-12-22T00:00:00",
        np.datetime64("2019-12-22T00:00:00"),
    ),
    ("http://www.w3.org/2001/XMLSchema#decimal", "10.10", 10.1),
]


class TestWoqlDataFrame:
    def test_extract_header_method(self, mock_result):
        df_header = woqlDF.extract_header(mock_result)
        assert df_header == [
            ("Product", "http://www.w3.org/2001/XMLSchema#string"),
            ("Qty", "http://www.w3.org/2001/XMLSchema#decimal"),
            ("Stock", "http://www.w3.org/2001/XMLSchema#string"),
            ("Warehouse", "http://www.w3.org/2001/XMLSchema#string"),
            ("buyers_buy_product", "http://www.w3.org/2001/XMLSchema#string"),
            ("stockDate", "http://www.w3.org/2001/XMLSchema#dateTime"),
            ("warehouseCode", "http://www.w3.org/2001/XMLSchema#string"),
        ]

    def test_extract_header_method_empty(self):
        with pytest.raises(EmptyException):
            woqlDF.extract_header({"bindings": []})

    def test_extract_column_method(self, mock_result):
        df_column = woqlDF.extract_column(
            mock_result, "Product", "http://www.w3.org/2001/XMLSchema#string"
        )
        assert df_column == ["STRAWBERRY CANDY"]

    @pytest.mark.parametrize("data_type,expected", TEST_TYPE)
    def test_type_map_method(self, data_type, expected):
        assert woqlDF.type_map(data_type) == expected

    def test_type_map_method_unknown(self):
        with pytest.raises(Exception, match=r"Unknown rdf type! .*"):
            woqlDF.type_map("some type that cause error")

    @pytest.mark.parametrize("data_type, value, expected", TEST_TYPE_VALUE)
    def test_type_value_map_method(self, data_type, value, expected):
        assert woqlDF.type_value_map(data_type, value) == expected

    def test_type_value_map_method_unknown(self):
        with pytest.raises(Exception, match=r"Unknown rdf type! .*"):
            woqlDF.type_value_map("some type that cause error", "some val")

    def test_query_to_df_method(self, mock_result):
        df = woqlDF.query_to_df(mock_result)
        assert df.shape == (1, 7)



WOQL_DELETE_JSON = {
    "@type": "woql:And",
    "woql:query_list": [
        {
            "@type": "woql:QueryListElement",
            "woql:index": {"@type": "xsd:nonNegativeInteger", "@value": 0},
            "woql:query": {
                "@type": "woql:DeleteQuad",
                "woql:subject": {"@type": "woql:Node", "woql:node": "scm:id"},
                "woql:predicate": {
                    "@type": "woql:Variable",
                    "woql:variable_name": {"@value": "Outgoing", "@type": "xsd:string"},
                },
                "woql:object": {
                    "@type": "woql:Variable",
                    "woql:variable_name": {"@value": "Value", "@type": "xsd:string"},
                },
                "woql:graph": {"@type": "xsd:string", "@value": "schema/main"},
            },
        },
        {
            "@type": "woql:QueryListElement",
            "woql:index": {"@type": "xsd:nonNegativeInteger", "@value": 1},
            "woql:query": {
                "@type": "woql:Optional",
                "woql:query": {
                    "@type": "woql:DeleteQuad",
                    "woql:subject": {
                        "@type": "woql:Variable",
                        "woql:variable_name": {
                            "@value": "Other",
                            "@type": "xsd:string",
                        },
                    },
                    "woql:predicate": {
                        "@type": "woql:Variable",
                        "woql:variable_name": {
                            "@value": "Incoming",
                            "@type": "xsd:string",
                        },
                    },
                    "woql:object": {"@type": "woql:Node", "woql:node": "scm:id"},
                    "woql:graph": {"@type": "xsd:string", "@value": "schema/main"},
                },
            },
        },
    ],
}



WOQL_CAST_JSON = {
    "@type": "woql:Typecast",
    "woql:typecast_value": {
        "@type": "woql:Variable",
        "woql:variable_name": {"@value": "Duration", "@type": "xsd:string"},
    },
    "woql:typecast_type": {"@type": "woql:Node", "woql:node": "xsd:integer"},
    "woql:typecast_result": {
        "@type": "woql:Variable",
        "woql:variable_name": {"@value": "Duration_Cast", "@type": "xsd:string"},
    },
}



from terminusdb_client.woqlclient.api_endpoint_const import APIEndpointConst

from .connectCapabilitiesResponse import ConnectResponse
from .getSchemaTurtleResponse import RESPONSE


def mocked_requests(*args, **kwargs):
    class MockResponse:
        def json(self):
            if self._json_data is None:
                raise ValueError("EXCEPTION NO JSON OBJECT")
            return self._json_data

        @property
        def status_code(self):
            return self._status_code

        @property
        def url(self):
            return self._url

        @property
        def text(self):
            return self._text

        def __init__(self, url, status, action_type):

            # set status code and content
            self._json_data = None
            self._text = None
            self._status_code = status
            self._content = "cont"
            self._url = url
            # add json data if provided

            if action_type == APIEndpointConst.CONNECT:
                self._json_data = ConnectResponse

            elif action_type == APIEndpointConst.GET_TRIPLES:
                self._text = RESPONSE

            # elif action_type == APIEndpointConst.WOQL_SELECT:
            #   with open("tests/getAllClassQueryResponse.json") as json_file:
            #      json_data = json.load(json_file)
            #     self._json_data = json_data
            #    json_file.close()

            elif (
                action_type == APIEndpointConst.CREATE_DATABASE
                or action_type == APIEndpointConst.DELETE_DATABASE
                or action_type == APIEndpointConst.UPDATE_TRIPLES
                or action_type == APIEndpointConst.BRANCH
                or action_type == APIEndpointConst.CREATE_GRAPH
            ):
                self._json_data = {"terminus:status": "terminus:success"}

    if (
        args[0]
        == "http://localhost:6363/branch/admin/myDBName/local/branch/my_new_branch"
    ):
        return MockResponse(args[0], 200, APIEndpointConst.BRANCH)

    elif (
        args[0]
        == "http://localhost:6363/graph/admin/myDBName/local/branch/master/instance/mygraph"
    ):
        return MockResponse(args[0], 200, APIEndpointConst.CREATE_GRAPH)

    elif (
        args[0]
        == "http://localhost:6363/triples/admin/myDBName/local/branch/master/instance/mygraph"
    ):
        return MockResponse(args[0], 200, APIEndpointConst.GET_TRIPLES)

    elif args[0] == "http://localhost:6363/db/admin/myFirstTerminusDB":
        return MockResponse(args[0], 200, APIEndpointConst.DELETE_DATABASE)

    return MockResponse(args[0], 200, APIEndpointConst.CONNECT)



import sys

from terminusdb_client.woqlclient.id_parser import IDParser

sys.path.append("woqlclient")


def test_parse_server_url():
    server_url = "http://localhost:6363/"
    id_parser = IDParser()
    assert id_parser.parse_server_url(server_url) == server_url


def test_parse_dbid():
    db_name = "myFirstTerminusDB"
    id_parser = IDParser()
    assert id_parser.parse_dbid(db_name) == db_name


def test_valid_url():
    full_url = "http://localhost:6363/myFirstTerminusDB"
    id_parser = IDParser()
    assert id_parser._valid_url(full_url)

    full_url = "localhost&899900/myFirstTerminusDB"
    assert not id_parser._valid_url(full_url)


def test_valid_id_string():
    id_string = "myFirstTerminusDB"
    id_parser = IDParser()
    assert id_parser._valid_id_string(id_string)
    id_string = "my First:"
    assert not id_parser._valid_id_string(id_string)



from .woql_query import WOQLQuery


class WOQLLib:

    """Patterns to help getting the most useful information from schema graphs
    """

    def __init__(self, mode=None):
        if mode is None:
            self._mode = "query"
        else:
            self._mode = mode
        self._user_variables = []
        self._user_values = []
        self._default_variables = []
        self._parameters = []
        self._default_schema_resource = "schema/main"
        self._default_commit_resource = "_commits"
        self._default_meta_resource = "_meta"
        self._masterdb_resource = "terminus"
        self._masterdb_doc = "terminus://terminus/data"
        self._empty = ""

    def classes(self, values, variables=None, schema_resource=None):
        """General Patten 1: Classes"""
        if schema_resource is not None:
            graph = schema_resource
        else:
            graph = self._default_schema_resource
        self._default_variables = [
            "Class ID",
            "Class Name",
            "Description",
            "Parents",
            "Children",
            "Abstract",
            "Parent",
            "Child",
        ]
        if variables:
            self._set_user_variables(variables)
        select_vars = self._get_varlist()[:6]
        qpattern = (
            WOQLQuery()
            .select(*select_vars)
            .woql_and(
                WOQLQuery().quad(self._varn("Class ID"), "type", "owl:Class", graph),
                WOQLQuery()
                .limit(1)
                .woql_or(
                    WOQLQuery().quad(
                        self._varn("Class ID"),
                        "comment",
                        self._varn("Description"),
                        graph,
                    ),
                    WOQLQuery().eq(self._varn("Description"), self._empty),
                ),
                WOQLQuery()
                .limit(1)
                .woql_or(
                    WOQLQuery().quad(
                        self._varn("Class ID"), "label", self._varn("Class Name"), graph
                    ),
                    WOQLQuery().eq(self._varn("Class Name"), ""),
                ),
                WOQLQuery()
                .limit(1)
                .woql_or(
                    WOQLQuery()
                    .quad(
                        self._varn("Class ID"),
                        "terminus:tag",
                        "terminus:abstract",
                        graph,
                    )
                    .eq(self._varn("Abstract"), "Yes"),
                    WOQLQuery().eq(self._varn("Abstract"), "No"),
                ),
                WOQLQuery()
                .limit(1)
                .woql_or(
                    WOQLQuery()
                    .group_by(
                        self._varn("Class ID"),
                        self._varn("Parent"),
                        self._varn("Parents"),
                    )
                    .woql_and(
                        WOQLQuery().quad(
                            self._varn("Class ID"),
                            "subClassOf",
                            self._varn("Parent"),
                            graph,
                        ),
                        WOQLQuery().woql_or(
                            WOQLQuery().eq(self._varn("Parent"), self._docroot()),
                            WOQLQuery().quad(
                                self._varn("Parent"), "type", "owl:Class", graph
                            ),
                        ),
                    ),
                    WOQLQuery().eq(self._varn("Parents"), ""),
                ),
                WOQLQuery()
                .limit(1)
                .woql_or(
                    WOQLQuery()
                    .group_by(
                        self._varn("Class ID"),
                        self._varn("Child"),
                        self._varn("Children"),
                    )
                    .woql_and(
                        WOQLQuery().quad(
                            self._varn("Child"),
                            "subClassOf",
                            self._varn("Class ID"),
                            graph,
                        ),
                        WOQLQuery().quad(
                            self._varn("Child"), "type", "owl:Class", graph
                        ),
                    ),
                    WOQLQuery().eq(self._varn("Children"), ""),
                ),
            )
        )
        return self._add_constraints(qpattern, values)

    def property(self, values, variables=None, schema_resource=None):
        """General Pattern 2: Properties"""
        if schema_resource is not None:
            graph = schema_resource
        else:
            graph = self._default_schema_resource
        self._default_variables = [
            "Property ID",
            "Property Name",
            "Property Domain",
            "Property Type",
            "Property Range",
            "Property Description",
            "OWL Type",
        ]
        if variables:
            self._set_user_variables(variables)
        select_vars = self._get_varlist[:6]
        qpattern = (
            WOQLQuery()
            .select(*select_vars)
            .woql_and(
                WOQLQuery()
                .quad(self._varn("Property ID"), "type", self._varn("OWL Type"), graph)
                .quad(
                    self._varn("Property ID"),
                    "domain",
                    self._varn("Property Domain"),
                    graph,
                )
                .quad(
                    self._varn("Property ID"),
                    "range",
                    self._varn("Property Range"),
                    graph,
                ),
                WOQLQuery()
                .limit(1)
                .woql_or(
                    WOQLQuery()
                    .eq(self._varn("OWL Type"), "owl:DatatypeProperty")
                    .eq(self._varn("Property Type"), "Data"),
                    WOQLQuery()
                    .eq(self._varn("OWL Type"), "owl:ObjectProperty")
                    .eq(self._varn("Property Type"), "Object"),
                ),
                WOQLQuery()
                .limit(1)
                .woql_or(
                    WOQLQuery().quad(
                        self._varn("Property ID"),
                        "label",
                        self._varn("Property Name"),
                        graph,
                    ),
                    WOQLQuery().eq(self._varn("Property Name"), ""),
                ),
                WOQLQuery()
                .limit(1)
                .woql_or(
                    WOQLQuery().quad(
                        self._varn("Property ID"),
                        "comment",
                        self._varn("Property Description"),
                        graph,
                    ),
                    WOQLQuery().eq(self._varn("Property Description"), ""),
                ),
            )
        )
        return self._add_constraints(qpattern, values)

    def graphs(self, values, variables=None, cresource=None):
        """General Pattern 3: Graphs
        Retrieves information about the graphs in existence at any commit (and whether the commit is the head of a branch)"""
        if cresource is None:
            cresource = self._default_commit_resource
        self._default_variables = [
            "Graph ID",
            "Graph Type",
            "Branch ID",
            "Commit ID",
            "Graph IRI",
            "Branch IRI",
            "Commit IRI",
        ]
        if variables:
            self._set_user_variables(variables)
        woql = WOQLQuery().woql_and(
            WOQLQuery().triple(
                self._varn("Graph IRI"), "ref:graph_name", self._varn("Graph ID")
            ),
            WOQLQuery().woql_or(
                WOQLQuery()
                .triple(
                    self._varn("Commit IRI"), "ref:instance", self._varn("Graph IRI")
                )
                .eq(self._varn("Graph Type"), "instance"),
                WOQLQuery()
                .triple(self._varn("Commit IRI"), "ref:schema", self._varn("Graph IRI"))
                .eq(self._varn("Graph Type"), "schema"),
                WOQLQuery()
                .triple(
                    self._varn("Commit IRI"), "ref:inference", self._varn("Graph IRI")
                )
                .eq(self._varn("Graph Type"), "inference"),
            ),
            WOQLQuery().triple(
                self._varn("Commit IRI"), "ref:commit_id", self._varn("Commit ID")
            ),
            WOQLQuery()
            .limit(1)
            .woql_or(
                WOQLQuery()
                .triple(
                    self._varn("Branch IRI"), "ref:ref_commit", self._varn("Commit IRI")
                )
                .triple(
                    self._varn("Branch IRI"), "ref:branch_name", self._varn("Branch ID")
                ),
                WOQLQuery()
                .eq(self._varn("Branch IRI"), "")
                .eq(self._varn("Branch ID"), ""),
            ),
        )
        compiled = self._add_constraints(woql, values)
        return WOQLQuery().using(cresource, compiled)

    def branches(self, values, variables=None, cresource=None):
        """General Pattern 4: Retrieves Branches, Their ID, Head Commit ID, Head Commit Time
        (if present, new branches have no commits)"""
        if cresource is None:
            cresource = self._default_commit_resource
        self._default_variables = [
            "Branch ID",
            "Time",
            "Commit ID",
            "Branch IRI",
            "Commit IRI",
        ]
        if variables:
            self._set_user_variables(variables)
        woql = (
            WOQLQuery()
            .triple(
                self._varn("Branch IRI"), "ref:branch_name", self._varn("Branch ID")
            )
            .opt()
            .triple(
                self._varn("Branch IRI"), "ref:ref_commit", self._varn("Commit IRI")
            )
            .triple(self._varn("Commit IRI"), "ref:commit_id", self._varn("Commit ID"))
            .triple(
                self._varn("Commit IRI"), "ref:commit_timestamp", self._varn("Time")
            )
        )
        compiled = self._add_constraints(woql, values)
        return WOQLQuery().using(cresource, compiled)

    def objects(self, values, variables=None):
        """General Pattern 5: Objects - just a list of object ids and their types"""
        self._default_variables = ["Object Type", "Object ID"]
        if variables:
            self._set_user_variables(variables)
        qpattern = WOQLQuery().triple(
            self._varn("Object ID"), "type", self._varn("Object Type")
        )
        return self._add_constraints(qpattern, values)

    def property_values(self, values, variables=None):
        """General Pattern 6: Full Object Properties"""
        self._default_variables = [
            "Object ID",
            "Property ID",
            "Property Value",
            "Value ID",
            "Value Class",
            "Any Class",
        ]
        if variables:
            self._set_user_variables(variables)
        select_vars = self._get_varlist()[:5]
        qpattern = (
            WOQLQuery()
            .select(*select_vars)
            .woql_or(
                WOQLQuery()
                .triple(
                    self._varn("Object ID"),
                    self._varn("Property ID"),
                    self._varn("Value ID"),
                )
                .triple(self._varn("Value ID"), "type", self._varn("Value Class"))
                .read_object(self._varn("Value ID"), self._varn("Property Value"))
                .comment(
                    "If the value of the property has a type, that means it is an object property and we retrieve the full object as the property value",
                ),
                WOQLQuery()
                .triple(
                    self._varn("Object ID"),
                    self._varn("Property ID"),
                    self._varn("Property Value"),
                )
                .woql_not()
                .triple(self._varn("Property Value"), "type", self._varn("Any Class"))
                .comment(
                    "If the value of the property has no type, that means it is a datatype property and we retrieve the datatype value as usual",
                ),
            )
        )

        return self._add_constraints(qpattern, values)

    def object_metadata(self, values, variables=None, schema_resource=None):
        """General Pattern 7: Object Metadata"""
        if schema_resource is not None:
            graph = schema_resource
        else:
            graph = self._default_schema_resource
        self._default_variables = [
            "Object ID",
            "Name",
            "Description",
            "Type ID",
            "Type Name",
            "Type Description",
        ]
        if variables:
            self._set_user_variables(variables)
        qpattern = (
            WOQLQuery()
            .triple(self._varn("Object ID"), "type", self._varn("Type ID"))
            .woql_and(
                WOQLQuery()
                .limit(1)
                .woql_or(
                    WOQLQuery().triple(
                        self._varn("Object ID"), "label", self._varn("Name")
                    ),
                    WOQLQuery().eq(self._varn("Name"), self._empty),
                ),
                WOQLQuery()
                .limit(1)
                .woql_or(
                    WOQLQuery().triple(
                        self._varn("Object ID"), "comment", self._varn("Description")
                    ),
                    WOQLQuery().eq(self._varn("Description"), self._empty),
                ),
                WOQLQuery()
                .limit(1)
                .woql_or(
                    WOQLQuery().quad(
                        self._varn("Type ID"), "label", self._varn("Type Name"), graph
                    ),
                    WOQLQuery().eq(self._varn("Type Name"), self._empty),
                ),
                WOQLQuery()
                .limit(1)
                .woql_or(
                    WOQLQuery().quad(
                        self._varn("Type ID"),
                        "comment",
                        self._varn("Type Description"),
                        graph,
                    ),
                    WOQLQuery().eq(self._varn("Type Description"), self._empty),
                ),
            )
            .comment(
                "Returns an object(s) with its id, name and descriptions, and the object type, type id and type description",
            )
        )
        return self._add_constraints(qpattern, values)

    def property_metadata(self, values, variables=None, schema_resource=None):
        """General Pattern 8: Value Metadata"""
        if schema_resource is not None:
            graph = schema_resource
        else:
            graph = self._default_schema_resource
        self._default_variables = [
            "Object ID",
            "Property ID",
            "Property Name",
            "Property Value",
            "Property Description",
        ]
        if variables:
            self._set_user_variables(variables)
        qpattern = (
            WOQLQuery()
            .triple(
                self._varn("Object ID"),
                self._varn("Property ID"),
                self._varn("Property Value"),
            )
            .limit(1)
            .woql_or(
                WOQLQuery().quad(
                    self._varn("Property ID"),
                    "label",
                    self._varn("Property Name"),
                    graph,
                ),
                WOQLQuery().eq(self._varn("Property Name"), self._empty),
            )
            .limit(1)
            .woql_or(
                WOQLQuery().quad(
                    self._varn("Property ID"),
                    "comment",
                    self._varn("Property Description"),
                    graph,
                ),
                WOQLQuery().eq(self._varn("Property Description"), self._empty),
            )
            .comment(
                "Returns a property and its value and the basic metadata associated with the property (name, id, description)",
            )
        )
        return self._add_constraints(qpattern, values)

    def commits(self, values, variables=None, cresource=None):
        """General Pattern 9: Commits"""
        if cresource is None:
            cresource = self._default_commit_resource
        self._default_variables = [
            "Commit ID",
            "Commit IRI",
            "Time",
            "Author",
            "Message",
            "Parent ID",
            "Parent IRI",
            "Children",
            "Child ID",
            "Child IRI",
            "Branch IRI",
            "Branch ID",
        ]
        if variables:
            self._set_user_variables(variables)
        qpattern = (
            WOQLQuery()
            .triple(self._varn("Commit IRI"), "ref:commit_id", self._varn("Commit ID"))
            .triple(
                self._varn("Commit IRI"), "ref:commit_timestamp", self._varn("Time")
            )
            .limit(1)
            .woql_or(
                WOQLQuery().triple(
                    self._varn("Commit IRI"), "ref:commit_author", self._varn("Author")
                ),
                WOQLQuery().eq(self._varn("Author"), self._empty),
            )
            .limit(1)
            .woql_or(
                WOQLQuery().triple(
                    self._varn("Commit IRI"),
                    "ref:commit_message",
                    self._varn("Message"),
                ),
                WOQLQuery().eq(self._varn("Message"), self._empty),
            )
            .limit(1)
            .woql_or(
                WOQLQuery()
                .triple(
                    self._varn("Commit IRI"),
                    "ref:commit_parent",
                    self._varn("Parent IRI"),
                )
                .triple(
                    self._varn("Parent IRI"), "ref:commit_id", self._varn("Parent ID")
                ),
                WOQLQuery()
                .eq(self._varn("Parent IRI"), self._empty)
                .eq(self._varn("Parent ID"), self._empty),
            )
            .limit(1)
            .woql_or(
                WOQLQuery()
                .select(self._varn("Children"))
                .group_by(
                    [self._varn("Commit IRI"), self._varn("Child IRI")],
                    self._varn("Child ID"),
                    self._varn("Children"),
                )
                .triple(
                    self._varn("Child IRI"),
                    "ref:commit_parent",
                    self._varn("Commit IRI"),
                )
                .triple(
                    self._varn("Child IRI"), "ref:commit_id", self._varn("Child ID")
                ),
                WOQLQuery().eq(self._varn("Children"), self._empty),
            )
            .limit(1)
            .woql_or(
                WOQLQuery()
                .triple(
                    self._varn("Branch IRI"), "ref:ref_commit", self._varn("Commit IRI")
                )
                .triple(
                    self._varn("Branch IRI"), "ref:branch_name", self._varn("Branch ID")
                ),
                WOQLQuery()
                .eq(self._varn("Branch IRI"), self._empty)
                .eq(self._varn("Branch ID"), self._empty),
            )
        )
        compiled = self._add_constraints(qpattern, values)
        return WOQLQuery().using(cresource, compiled)

    def commit_chain(self, values, variables=None, cresource=None):
        """General Pattern 10: Commit Chain"""
        if cresource is None:
            cresource = self._default_commit_resource
        self._default_variables = ["Head IRI", "Tail IRI", "Path"]
        if variables:
            self._set_user_variables(variables)
        woql = WOQLQuery().path(
            self._varn("Head IRI"),
            "ref:commit_parent+",
            self._varn("Tail IRI"),
            self._varn("Path"),
        )
        compiled = self._add_constraints(woql, values)
        return WOQLQuery().using(cresource, compiled)

    def repos(self, values, variables=None, cresource=None):
        """General Pattern 11: Repositories"""
        if cresource is None:
            cresource = self._default_commit_resource
        self._default_variables = [
            "Repository IRI",
            "Repository Name",
            "Repository Type",
            "Remote URL",
            "Repository Head IRI",
        ]
        if variables:
            self._set_user_variables(variables)
        woql = WOQLQuery().woql_and(
            WOQLQuery()
            .triple(
                self._varn("Repository IRI"),
                "repo:repository_name",
                self._varn("Repository Name"),
            )
            .triple(self._varn("Repository IRI"), "type", self._varn("Repository Type"))
            .triple(
                self._varn("Repository IRI"),
                "repo:repository_head",
                self._varn("Repository Head IRI"),
            )
            .woql_or(
                WOQLQuery().eq(self._varn("Repository Type"), "repo:Local"),
                WOQLQuery()
                .eq(self._varn("Repository Type"), "repo:Remote")
                .triple(
                    self._varn("Repository IRI"),
                    "repo:remote_url",
                    self._varn("Remote URL"),
                ),
            )
        )
        compiled = self._add_constraints(woql, values)
        return WOQLQuery().using(cresource, compiled)

    def dbs(self, values, variables=None):
        """General Pattern 12: dbs"""
        self._default_variables = [
            "DB Name",
            "DB ID",
            "Account",
            "Resource Name",
            "Description",
            "Allow Origin",
            "DB IRI",
            "DB x",
        ]
        if variables:
            self._set_user_variables(variables)
        woql = WOQLQuery().woql_and(
            WOQLQuery()
            .triple(self._varn("DB IRI"), "type", "terminus:Database")
            .triple(
                self._varn("DB IRI"),
                "terminus:resource_name",
                self._varn("Resource Name"),
            )
            .triple(
                self._varn("DB IRI"),
                "terminus:allow_origin",
                self._varn("Allow Origin"),
            )
            .split(
                self._varn("Resource Name"),
                "\\|",
                [self._varn("Account"), self._varn("DB ID")],
            )
            .woql_and(
                WOQLQuery()
                .limit(1)
                .woql_or(
                    WOQLQuery().triple(
                        self._varn("DB IRI"), "label", self._varn("DB Name")
                    ),
                    WOQLQuery().eq(self._varn("DB Name"), self._empty),
                ),
                WOQLQuery()
                .limit(1)
                .woql_or(
                    WOQLQuery().triple(
                        self._varn("DB IRI"), "comment", self._varn("Description")
                    ),
                    WOQLQuery().eq(self._varn("Description"), self._empty),
                ),
            )
        )
        compiled = self._add_constraints(woql, values)
        return WOQLQuery().using(self.masterdb_resource, compiled)

    def users(self, values, variables=None):
        """General Pattern 13: users"""
        self._default_variables = [
            "User ID",
            "User Name",
            "Commit Log ID",
            "User Notes",
            "Capabilities",
            "User IRI",
            "Capability IRI",
        ]
        if variables:
            self._set_user_variables(variables)
        select_vars = self._get_varlist()[:6]

        woql = (
            WOQLQuery()
            .select(*select_vars)
            .woql_and(
                WOQLQuery()
                .triple(self._varn("User IRI"), "type", "terminus:User")
                .triple(
                    self._varn("User IRI"), "terminus:agent_name", self._varn("User ID")
                )
                .woql_and(
                    WOQLQuery()
                    .limit(1)
                    .woql_or(
                        WOQLQuery().triple(
                            self._varn("User IRI"), "label", self._varn("User Name")
                        ),
                        WOQLQuery().eq(self._varn("User Name"), self._empty),
                    ),
                    WOQLQuery()
                    .limit(1)
                    .woql_or(
                        WOQLQuery().triple(
                            self._varn("User IRI"),
                            "terminus:commit_log_id",
                            self._varn("Commit Log ID"),
                        ),
                        WOQLQuery().eq(self._varn("Commit Log ID"), self._empty),
                    ),
                    WOQLQuery()
                    .limit(1)
                    .woql_or(
                        WOQLQuery().triple(
                            self._varn("User IRI"), "comment", self._varn("User Notes")
                        ),
                        WOQLQuery().eq(self._varn("User Notes"), self._empty),
                    ),
                    WOQLQuery()
                    .limit(1)
                    .woql_or(
                        WOQLQuery()
                        .group_by(
                            self._varn("User IRI"),
                            self._varn("Capability IRI"),
                            self._varn("Capabilities"),
                        )
                        .triple(
                            self._varn("User IRI"),
                            "terminus:authority",
                            self._varn("Capability IRI"),
                        ),
                        WOQLQuery().eq(self._varn("Capabilities"), self._empty),
                    ),
                )
            )
        )
        compiled = self._add_constraints(woql, values)
        return WOQLQuery().using(self.masterdb_resource, compiled)

    def capabilities(self, values, variables=None):
        """General Pattern 14: capabilities"""
        self._default_variables = [
            "Capability Name",
            "Capability Type",
            "Capability Description",
            "Capability IRI",
            "Access IRI",
            "Permissions",
            "Resources",
            "Permission ID",
            "Resource ID",
        ]
        if variables:
            self._set_user_variables(variables)
        # select_vars = self._get_varlist()[:7]
        woql = WOQLQuery().woql_and(
            WOQLQuery()
            .triple(self._varn("Capability IRI"), "type", self._varn("Capability Type"))
            .sub("terminus:Capability", self._varn("Capability Type")),
            WOQLQuery()
            .limit(1)
            .woql_or(
                WOQLQuery().triple(
                    self._varn("Capability IRI"), "label", self._varn("Capability Name")
                ),
                WOQLQuery().eq(self._varn("Capability Name"), self.empty),
            ),
            WOQLQuery()
            .limit(1)
            .woql_or(
                WOQLQuery().triple(
                    self._varn("Capability IRI"),
                    "comment",
                    self._varn("Capability Description"),
                ),
                WOQLQuery().eq(self._varn("Capability Description"), self.empty),
            ),
            WOQLQuery()
            .triple(
                self._varn("Capability IRI"),
                "terminus:access",
                self._varn("Access IRI"),
            )
            .group_by(
                self._varn("Access IRI"),
                self._varn("Permission ID"),
                self._varn("Permissions"),
            )
            .triple(
                self._varn("Access IRI"), "terminus:action", self._varn("Permission ID")
            ),
            WOQLQuery()
            .triple(
                self._varn("Capability IRI"),
                "terminus:access",
                self._varn("Access IRI"),
            )
            .group_by(
                self._varn("Access IRI"),
                self._varn("Resource ID"),
                self._varn("Resources"),
            )
            .triple(
                self._varn("Access IRI"),
                "terminus:authority_scope",
                self._varn("Resource ID"),
            ),
        )
        compiled = self._add_constraints(woql, values)
        return WOQLQuery().using(self._masterdb_resource, compiled)

    def add_user(self, user_id, values, variables=None):
        doc = self.masterdb_doc
        self.default_variables = [
            "User Name",
            "User Commit ID",
            "Notes",
            "Password",
            "User ID",
            "User IRI",
        ]
        if variables:
            self._set_user_variables(variables)
        query = (
            WOQLQuery()
            .idgen(doc + "User", user_id, self._varn("User IRI"))
            .add_triple(self._varn("User IRI"), "type", "terminus:User")
            .add_triple(self._varn("User IRI"), "terminus:agent_name", user_id)
            .add_triple(self._varn("User IRI"), "label", self._varn("User Name"))
            # .add_triple(self._varn('User IRI'), 'terminus:password', self._varn('Password'))
            .add_triple(
                self._varn("User IRI"),
                "terminus:commit_log_id",
                self._varn("User Commit ID"),
            )
            .add_triple(self._varn("User IRI"), "comment", self._varn("Notes"))
        )
        woql = self._add_constraints(query, values)
        return WOQLQuery().using(self._masterdb_resource, woql)

    def add_capability(self, user_id, label=None, description=None):
        query = WOQLQuery().add_triple(id, "type", "terminus:Capability")
        if label:
            query.add_triple(user_id, "label", label)
        if description:
            query.add_triple(user_id, "comment", description)
        return WOQLQuery().using(self.masterdb_resource, query)

    def add_access(self, accid, permissions, resources, label=None, description=None):
        query = WOQLQuery()
        if not accid:
            accid = "v:Access IRI"
            query.when(
                WOQLQuery().idgen("doc:Access", [permissions.concat(resources)], accid)
            )
        query.add_triple(accid, "type", "terminus:Access")
        for item in permissions:
            query.add_triple(accid, "terminus:action", item)
        for item in resources:
            query.add_triple(accid, "terminus:authority_scope", item)
        if label:
            query.add_triple(accid, "label", label)
        if description:
            query.add_triple(accid, "comment", description)
        return WOQLQuery().using(self.masterdb_resource, query)

    def grant_access(self, capiri, acciri):
        query = WOQLQuery().add_triple(capiri, "terminus:access", acciri)
        return WOQLQuery().using(self.masterdb_resource, query)

    def grant_capability(self, uid, capiri, uiri=None):
        if uiri is None:
            uiri = "v:User IRI"
        query = (
            WOQLQuery()
            .when(WOQLQuery().triple(uiri, "terminus:agent_name", uid))
            .add_triple(uiri, "terminus:authority", capiri)
        )
        return WOQLQuery().using(self.masterdb_resource, query)

    def revoke_capability(self, uid, capiri, uiri=None):
        if uiri is None:
            uiri = "v:User IRI"
        query = (
            WOQLQuery()
            .when(WOQLQuery().triple(uiri, "terminus:agent_name", uid))
            .delete_triple(uiri, "terminus:authority", capiri)
        )
        return WOQLQuery().using(self.masterdb_resource, query)

    def prefixes(self, values, variables=None, cresource=None):
        if cresource is None:
            cresource = self._default_commit_resource
        self.default_variables = ["Prefix", "URI", "Prefix Pair IRI"]
        if variables:
            self._set_user_variables(variables)
        query = (
            WOQLQuery()
            .triple(
                "ref:default_prefixes", "ref:prefix_pair", self._varn("Prefix Pair IRI")
            )
            .triple(self._varn("Prefix Pair IRI"), "ref:prefix_uri", self._varn("URI"))
            .triple(self._varn("Prefix Pair IRI"), "ref:prefix", self._varn("Prefix"))
        )
        woql = self._add_constraints(query, values)
        return WOQLQuery().using(cresource, woql)

    def insert_prefix(self, values, variables=None, cresource=None):
        if cresource is None:
            cresource = self._default_commit_resource
        self.default_variables = ["Prefix Pair IRI"]
        if variables:
            self._set_user_variables(variables)
        query = (
            WOQLQuery()
            .using(cresource)
            .idgen("doc:PrefixPair", values, self._varn("Prefix Pair IRI"))
            .add_triple(
                "ref:default_prefixes", "ref:prefix_pair", self._varn("Prefix Pair IRI")
            )
            .insert(self._varn("Prefix Pair IRI"), "ref:PrefixPair")
            .property("ref:prefix", values[0])
            .property("ref:prefix_uri", values[1])
        )
        return query

    def document_classes(self, values, variables=None, graph_resource=None):
        pattern = self.classes(values, variables, graph_resource)
        return self._add_doctype_constraint(pattern, self._varn("Class ID"))

    def user_capabilities(self, values, variables=None):
        self.capabilities(False, variables)  # load default variables
        self._default_variables.append("User IRI")
        self.default_variables.append("User ID")
        if variables:
            self._set_user_variables(variables)
        constraint = (
            WOQLQuery()
            .triple(
                self._varn("User IRI"), "terminus:agent_name", self._varn("User ID")
            )
            .triple(
                self._varn("User IRI"),
                "terminus:authority",
                self._varn("Capability IRI"),
            )
        )
        return WOQLLib().capabilities(constraint, variables)

    def commit_chain_full(self, values, graph_resource=None):
        if graph_resource is None:
            cresource = self._default_commit_resource
        else:
            cresource = graph_resource
        woql = WOQLQuery().woql_or(
            WOQLQuery().woql_and(
                WOQLLib().commit_chain(None, ["Commit IRI"]), WOQLLib().commits()
            ),
            WOQLQuery().woql_and(
                WOQLLib().commit_chain(None, ["Head IRI", "Commit IRI"]),
                WOQLLib().commits(),
            ),
        )
        compiled = self._add_constraints(woql, values)
        return WOQLQuery().using(cresource, compiled)

    def concrete_document_classes(self, values, variables=None, graph_resource=None):
        pattern = self.document_classes(values, variables, graph_resource)
        if graph_resource is None:
            graph = self._default_schema_resource
        else:
            graph = graph_resource
        return self._add_constraint(
            pattern,
            WOQLQuery()
            .woql_not()
            .quad(
                self._varn("Document IRI"), "terminus:tag", "terminus:abstract", graph
            )
            .comment("Does not match document class(es) marked as abstract"),
        )

    def document_metadata(self, values, variables=None, graph_resource=None):
        if not variables:
            variables = []
        pattern = self.object_metadata(values, variables, graph_resource)
        return self._add_doctype_constraint(pattern, self._varn("Type ID"))

    def documents(self, values, variables=None):
        if not variables:
            variables = ["Document Type", "Document ID"]
        pattern = self.objects(values, variables)
        return self._add_constraints(pattern, values)

    def first_commit(self):
        pattern = self.commits()
        noparent = (
            WOQLQuery()
            .using("_commits")
            .woql_and(
                WOQLQuery().triple(
                    self._varn("Any Commit IRI"),
                    "ref:commit_id",
                    self._varn("Commit ID"),
                ),
                WOQLQuery()
                .woql_not()
                .triple(
                    self._varn("Any Commit IRI"),
                    "ref:commit_parent",
                    self._varn("Parent IRI"),
                ),
            )
        )
        return self._add_constraint(pattern, noparent)

    def _add_constraints(self, pattern, vals):
        if not vals:
            return pattern
        if isinstance(vals, list):
            nq = WOQLQuery()
            for idx, item in enumerate(vals):
                if item is not None:
                    nq.woql_and(WOQLQuery().eq(item, self._var(idx)))
            return self._add_constraint(pattern, nq)
        elif hasattr(vals, "json"):
            return self._add_constraint(pattern, vals)
        elif isinstance(vals, dict):
            myvars = self._get_varlist()
            nq = WOQLQuery()
            for key, item in vals.items():
                if key in myvars:
                    nq.woql_and(WOQLQuery().eq(item), self._var(myvars.index(key)))
            return self._add_constraint(pattern, nq)
        else:
            return self._add_constraint(pattern, WOQLQuery().eq(self._var(0), vals))

    def _add_constraint(self, qpattern, woql):
        amalgamated = WOQLQuery().woql_and(woql, qpattern)
        return amalgamated

    def _var(self, idx):
        if idx >= len(self._user_variables):
            vname = self._default_variables[idx]
        else:
            vname = self._user_variables[idx]
        return "v:" + vname

    def _varn(self, default_varname):
        return self._var(self._default_variables.index(default_varname))

    def _get_varlist(self):
        varlist = []
        for idx, item in enumerate(self._default_variables):
            if idx < len(self._user_variables) and self._user_variables[idx]:
                varlist.append(self._user_variables[idx])
            else:
                varlist.append(item)
        return varlist

    def _get_meta(self, query):
        meta = {
            "variables": self._get_varlist(),
            "default_variables": self.default_variables,
            "query": query.to_dict(),
        }
        if self._user_variables:
            meta._user_variables = self._user_variables
        if self._user_values:
            meta._user_values = self._user_value
        return meta

    def _docroot(self):
        return "terminus:Document"

    def _add_doctype_constraint(self, pattern, varname):
        self._add_constraint(
            pattern,
            WOQLQuery()
            .sub(self._docroot(), varname)
            .comment("Only matches classes subsumed by terminus:Document class"),
        )
